# This file contains some functions that deal with the manifold's cusps.
# It contains in particular the functions that compute the cusp shape x + iy
# and the maximal cusp area A (both functions take errors into account)

from string import split
from imath import iabs, isqrt#, ilog#, iexp
# from math import sqrt, log, exp
# import cmath
import hikmot.interval as itl
import hikmot.complex as ic
# import os
# import snappy
from sys import float_info
# from time import sleep
from hikmot import verify_hyperbolicity

# constants.py contains the global constants used in all files

from constants import *

# Calculates the maximum area of a horoball section on a cusp neighborhood (all sections are tied).
# The function uses the SnapPea Cusp Neighborhood function. If it fails, it returns -1.

def maximal_area(N):
	M = N.filled_triangulation()
	try:
		CN = M.cusp_neighborhood()
		num_cusps = CN.num_cusps()

		# Ties all cusps

		for i in range(num_cusps):
			CN.set_tie(i,1)

		# Maximises the horocusp section. The level 'max_reach' gives the maximum reachable
		# by a single cusp and is thus in general bigger than what we need here, so it's fine.

		CN.set_displacement(CN.max_reach())	
		volume = CN.volume()

		# Area = twice the volume!

		area = 2*volume
		return area

	# If SnapPea is not able to construct CN, return -1
	
	except RuntimeError:
		if constants.verbose>=2:
			print 'failed constructing cusp neighborhoods'
		return -1

# Returns the shape (x,y) of the cusp number 'cusp' (starting from zero, as usual),
# as computed by SnapPea

def cusp_shape(N, cusp):
	c = N.cusp_info(cusp)['shape']
	x = c.real
	y = c.imag
	return (x,y)

# Returns the triangulation of the manifold. The format is
# triangulation[num_tetrahedron][num_face] = [2, [0, 3, 2, 1], 1] 
# following SnapPea convention:
# "2" is the neighbor tetrahedron
# "[0, 3, 2, 1]" tells you how to glue the two tetrahedra
# "1" is the cusp number of the vertex
# This information is found (inefficiently)
# by saving the manifold on a file and then reading it.

def get_triangulation(N):
	num_cusps = N.num_cusps()
	num_tetrahedra = N.num_tetrahedra()

	# Import data on triangulation and store it in lines:
	
	N.save('temp.tri')
	f = open('temp.tri', 'r') 
	lines = f.readlines()
	l = 9 + num_cusps	
	triangulation = []

	# From lines to the triangulation list:
	
	for i in range(num_tetrahedra):
		line_a = lines[l]
		words_a = split(line_a)
		line_b = lines[l+1]
		words_b = split(line_b)
		line_c = lines[l+2]
		words_c = split(line_c)
		new_element = []
		for j in range(4):
			new_element.append([0, [0, 0, 0, 0], 0])
			new_element[j][0] = int(words_a[j])			
			for k in range(4):
				new_element[j][1][k] = int(words_b[j][k])
			new_element[j][2] = int(words_c[j])
		triangulation.append(new_element[:])
		l = l + 9
	f.close()
	return triangulation		

# Returns a list cusp[tet][vertex] = 
# 	number of the cusp containing the vertex 'vertex' of the tetrahedron 'tet'
# This function is actually not used by our program!

def get_cusp(triangulation):
	num_tetrahedra = len(triangulation)
	cusp = []

	# Initialize cusps = [[0, 1, 2, 3], [4, 5, 6, 7], ... ]

	for i in range(num_tetrahedra):
		start = range(4*i, 4*(i+1))
		cusp.append(start[:])

	# Whenever two incident tetrahedra identify pairs of vertices of 
	# distinct labels a and b, with (say) a < b,
	# change all labels b into a.
	# Then repeat as much as you can

	keep_going_on = True
	while keep_going_on == True:
		keep_going_on = False
		for tet in range(num_tetrahedra):
			for vertex in range(4):
				a = cusp[tet][vertex]
				for exit in range(4):
					if exit != vertex:
						companion_tet = triangulation[tet][exit][0]
						companion_vertex = triangulation[tet][exit][1][vertex]
						b = cusp[companion_tet][companion_vertex]						
						if a < b:
							
							# Substitutes all labels b with a
							# and decrease by one all labels > b
							
							for t in range(num_tetrahedra):
								for v in range(4):
									if cusp[t][v] == b:
										cusp[t][v] = a
									if cusp[t][v] > b:
										cusp[t][v] = cusp[t][v] - 1
							
							keep_going_on = True					
	return cusp

# Returns a euclidean triangulation of the cusps.
# edge_length[tet][vertex][edge] is the complex length of the edge 'edge'.
# in the triangle section at vertex 'vertex' of the tetrahedron 'tet'
# It starts by assigning value "1" to one arbitrary edge on each cusp.
# The input tetrahedra_shapes are intervals and the output lengths also are.

def get_euclidean_triangulation(N, triangulation, tetrahedra_shapes):

	# Initialize various informations on the triangulation
	
	num_tetrahedra = len(triangulation)
	num_tetrahedra = len(triangulation)
	num_cusps = N.num_cusps()
	
	# Initialize edge_length

	edge_length = []
	for tet in range(num_tetrahedra):
		new_element = []
		for vertex in range(4):
			new_element.append([0, 0, 0, 0])
		edge_length.append(new_element[:])

	# On each cusp, set one arbitrary edge_length to 1
		
	cusp_done = []
	for i in range(num_cusps):
		cusp_done.append(False)
	for tet in range(num_tetrahedra):
		for vertex in range(4):
			which_cusp = triangulation[tet][vertex][2]
			if cusp_done[which_cusp] == False:
				edge = (vertex + 1) % 4
				edge_length[tet][vertex][edge] = ic.complex(1,0)
				cusp_done[which_cusp] = True
					
	# Sets all complex lengths using the tetrahedra shape
	# TBD: Make a more efficient routine (if possible)
	
	keep_going_on = True
	while keep_going_on == True:
		keep_going_on = False
		for tet in range(num_tetrahedra):
			shape = tetrahedra_shapes[tet]
			for vertex in range(4):
				which_cusp = triangulation[tet][vertex][2]
				for edge in range(4):
					if edge != vertex and edge_length[tet][vertex][edge] != 0:

						# Set the left edge in the triangle

						left_vertex, right_vertex = get_oriented_vertices(edge, vertex)
						left_edge = right_vertex
						if edge_length[tet][vertex][left_edge] == 0:
							complex_angle = get_edge_shape ((vertex, left_vertex), shape)
							edge_length[tet][vertex][left_edge] = - edge_length[tet][vertex][edge] * complex_angle
							keep_going_on = True

						# Set the corresponding edge on the adjacent tetrahedron

						adjacent_tet = triangulation[tet][edge][0]
						adjacent_vertex = triangulation[tet][edge][1][vertex]
						adjacent_edge = triangulation[tet][edge][1][edge]
						if edge_length[adjacent_tet][adjacent_vertex][adjacent_edge] == 0:
							edge_length[adjacent_tet][adjacent_vertex][adjacent_edge] = \
								- edge_length[tet][vertex][edge]

							keep_going_on = True					
	return edge_length

# Check that the cusp section determined by 'edge_length' exists
# That is, it checks that each euclidean triangle is small enough to be embedded in its tetrahedron
# If this is not the case, the function returns a rescaling factor, so that
# after rescaling by that factor the cusp section exists.

def check_cusp_section_exists(N, triangulation, edge_length, tetrahedra_shapes):
	num_tetrahedra = len(triangulation)
	check = True
	rescaling_factor = itl.interval(1.0)
	for tet in range(num_tetrahedra):
		
		# Put the tetrahedron in half-space with vertices at 0, 1, z, infinity.
		# The 'height' here is the minimum level h such that the plane x_3 = h 
		# intersects the tetrahedron in a triangle. (x_1, x_2, x_3 are variables for three-space.)
		#
		# Let C be the circle containing 0, 1, z. Our tetrahedron is the intersection
		# of four half-spaces, three vertical ones and a fourth one with boundary at C.
		# If the center of C is contained in the triangle T with vertices 0, 1, z, then height = radius of C
		# If not, then height = (length of the longest edge in T) / 2 = (max (1, |z|, |z-1|) / 2)
		
		z = tetrahedra_shapes[tet]
		height = 0
				
		# If the center of C is not contained in T:

		# Modified 5 may 2014. We substitute the following:
		#
		# if z.real.inf < 0 or z.real.sup > 1 or iabs(z-0.5).inf > 0.5:
		#	height = max(1, iabs(z-1), iabs(z))/2

		if z.real.sup < 0 or z.real.inf > 1 or iabs(z-0.5).inf > 0.5:
			height = itl.interval(max(1, iabs(z-1).sup, iabs(z).sup)/2)

		# If the center of C is contained in T, i.e. if internal angles are < 90 degrees:

		else:
			a = z.real
			b = z.imag

			# compute height

			height = isqrt(b*b + (a*a + b*b -a)*(a*a + b*b -a)) /2 /b	

		for vertex in range(4):
			companion_vertex = [1, 0, 3, 2]
			other_vertex = companion_vertex[vertex]

			length = iabs(edge_length[tet][vertex][other_vertex])

			# Here triangle_height = abs(z - 1) / length.
			
			triangle_height = (iabs(z - 1)) / length
			if triangle_height.sup <= height.inf and triangle_height/height.sup < rescaling_factor.sup :
				check = False
				rescaling_factor = triangle_height/height
	return check, rescaling_factor

# Returns a list of areas of all cusp sections

def cusps_area(triangulation, num_cusps, edge_length):
	num_tetrahedra = len(triangulation)
	area = []
	for i in range(num_cusps):
		area.append(0)
	for tet in range(num_tetrahedra):
		for vertex in range(4):
			cusp = triangulation[tet][vertex][2]

			# Use Heron's formula to compute the area of the euclidean triangle

			p = itl.interval(0)

			for edge in range(4):
				if edge != vertex:
					p = p + iabs(edge_length[tet][vertex][edge])
			area_triangle = p/2
			for edge in range(4):
				if edge != vertex:
					factor = p/2 - iabs(edge_length[tet][vertex][edge])
					area_triangle = area_triangle * factor
			area_triangle = isqrt(area_triangle)

			area[cusp] = area[cusp] + area_triangle

	return area[:]

# Returns the (exponential of the) minimum distance (along the edges of the triangulation) of 
# cusp sections. If the triangulation is canonical, this is the minimum distance
# between distinct horoball lifts of the sections.
# Takes errors into account: it returns a number which is certainly smaller than the true minimum distance.
# The function needs canonization. If canonization fails, it returns -1.

def get_min_distance(triangulation, edge_length):
	num_tetrahedra = len(triangulation)
	min_distance = itl.interval(-1)

	# See SnapPea for the (simple) formula we use here: d = -0.5 * log (e1 * e2 * e1' * e2')
	# Where e1 .. e2' are the lengths of the four (triangle) edges that are incident to the (tetrahedron) edge
	# Since we use the exponential, we get distance = e^d = 1/sqrt (e1 * e2 * e1' * e2')

	for tet in range(num_tetrahedra):
		for vertex_a in range(4):
			for vertex_b in range(vertex_a + 1, 4):
				distance = itl.interval(1)
				for v in [vertex_a, vertex_b]:
					for j in range(4):
						if j != vertex_a and j != vertex_b:
							factor = iabs(edge_length[tet][v][j])
							distance = distance * factor
				distance = 1/isqrt(distance)
				if (min_distance.inf <  0 or distance.inf < min_distance.inf):
					min_distance = distance
	return min_distance



# Returns the area of any cusp in the maximal cusp section where all are cusps are tied. 
# If it fails to find this area, it returns -1
# Otherwise it returns area, N, tetrahedra_shapes where N is the canonized triangulation
# The returned N is hikmot-verified.

def get_maximal_cusp_section(M):
	try:
		N = M.filled_triangulation()
		N.canonize()

		# Hyperbolicity of the canonized triangulation needs to be verified 

		if verify_with_hikmot:
			flag, tetrahedra_shapes = verify_hyperbolicity(N,verbose >=2)
			if flag == False:	
				return -1, N, 0

		# Initialize various informations on the triangulation
		
		triangulation = get_triangulation(N)
		num_tetrahedra = len(triangulation)
		num_cusps = N.num_cusps()
		
		# Construct a euclidean triangulation of the cusps

		edge_length = get_euclidean_triangulation(N, triangulation, tetrahedra_shapes)

		# Calculate the area of each cusp section		

		area = cusps_area(triangulation, num_cusps, edge_length)

		# Normalize all lengths in order to have all areas equal to (3 * sqrt(3))/8

		fixed_area = (3 * isqrt(3))/8 	# = 0.649519052838...

		for tet in range(num_tetrahedra):
			for vertex in range(4):
				which_cusp = triangulation[tet][vertex][2]
				normalizing_factor = ic.complex(isqrt(fixed_area / (area[which_cusp])))
				for edge in range(4):
					if edge != vertex:
						edge_length[tet][vertex][edge] = \
							edge_length[tet][vertex][edge] * normalizing_factor

		# Compute the minimum distance along edges of the triangulation
		# between the two sections. Takes errors into account: the function returns a number
		# which is certainly smaller than the real one (so cusp sections will certainly not overlap)

		min_distance = get_min_distance(triangulation, edge_length)

		# Recall that min_distance is actually the exponential of the distance

		final_area = fixed_area * min_distance

		# Re-normalize all lengths to get the maximal section.

		normalizing_factor = ic.complex(isqrt(final_area / fixed_area))

		for tet in range(num_tetrahedra):
			for vertex in range(4):
				for edge in range(4):
					if edge != vertex:
						edge_length[tet][vertex][edge] = \
							edge_length[tet][vertex][edge] * normalizing_factor

		# Check that this section indeed exists (i.e. on each tetrahedron the euclidean triangles
		# can indeed be constructed, because they are not too big). The answer should be positive since
		# the triangulation is canonical (or very near to a canonical one).

		flag, rescaling_factor = check_cusp_section_exists(N, triangulation, edge_length, tetrahedra_shapes)
		if flag == False:
			
			# Use a slightly smaller factor 

			normalizing_factor = ic.complex(rescaling_factor-0.01)
			for tet in range(num_tetrahedra):
				for vertex in range(4):						
					for edge in range(4):
						if edge != vertex:
							edge_length[tet][vertex][edge] = \
							    edge_length[tet][vertex][edge] * normalizing_factor

			final_area = final_area * (normalizing_factor * normalizing_factor)

			# If for some reason it still does not pass the check, we give up

			if check_cusp_section_exists(N, triangulation, edge_length, tetrahedra_shapes)[0] == False:
				if verbose:
					print 'I could not construct a cusp section'
					return -1

		return final_area, N, tetrahedra_shapes

	# If SnapPea is not able to canonize, return -1
	
	except RuntimeError:
		return -1, N, 0
		
# Returns a list peripheral[which_curve][tet][vertex] = [a, b, c, d] which encodes
# the peripheral curve as in Snappea file format.
# The curve is the meridian/longitude depending on which_curve = 0/1.
#
# Recall that SnapPea format is as follows:
# The peripheral curve is in normal position with respect to the triangulation
# of the cusp induced by the triangulation of N. The curve is oriented.
# It may intersect one edge multiple times, with the same direction.
# For every vertex of tet we have a triangle in some cusp. 
# The number in [a, b, c, d] in position 'vertex' is always zero and carries no info.
# The other three integers indicate (in absolute value) the intersection numbers
# with the three edges of the triangle. The sign indicates whether the curve enters
# or exit the triangle: their sum on each triangle is therefore always zero.

def get_peripheral_curves(N):
	N.save('temp.tri')
	f = open('temp.tri', 'r') 
	lines = f.readlines()
	peripheral = [[], []]
	for which_curve in range(2):
		l = 12 + N.num_cusps()
		if which_curve == 1: l = l + 2
		for tet in range(N.num_tetrahedra()):
			new_element = []
			line = lines[l]
			words = split(line)
			for vertex in range(4):
				new_element.append([0, 0, 0, 0])
				for i in range(4):
					new_element[vertex][i] = int(words[4*vertex + i])
			peripheral[which_curve].append(new_element[:])
			l = l + 9
	f.close()
	return peripheral	

# Finds an edge in the triangulation of the cusp 'num_cusp' which intersects
# both the meridian and longitude.
# The edge is encoded as (num_tetrahedron, num_vertex, num_edge)
# The pair (num_tetrahedron, num_vertex) determines a triangle in the cusp,
# whose vertices (edges) are numbered according to the closer (opposite) vertex
# of the tetrahedron.
#
# Takes as arguments the lists 'triangulation', 'peripheral_curves', 'cusp' 
# obtained from get_triangulation(), get_peripheral_curves(), get_cusp()

def get_starting_edge(triangulation, peripheral_curves, num_cusp):
	num_tetrahedra = len(triangulation)
	for tet	in range(num_tetrahedra):
		for vertex in range(4):
			if triangulation[tet][vertex][2] == num_cusp:
				for edge in range(4):
					if edge != vertex:
						intersects = True
						for which_curve in range(2):
							if peripheral_curves[which_curve][tet][vertex][edge] == 0:
								intersects = False
						if intersects == True:
							return (tet, vertex, edge)
	return "Error"

# The number i is an element in the set {0,1,2,3} \ {miss}, which is cyclically
# ordered. The function returns the number after i following this ordering.

def add_one (i, miss):
	j = (i+1) % 4
	if j != miss: return j
	k = (j+1) % 4
	return k

# Returns the pair (left_vertex, right_vertex) of the oriented edge 'edge' 
# in the oriented triangular cusp section of a tetrahedron (at vertex 'vertex')

def get_oriented_vertices(edge, vertex):
	left_vertex = add_one(edge, vertex)
	right_vertex = add_one(left_vertex, vertex)
	if vertex % 2 == 0:
		return (left_vertex, right_vertex)
	else:
		return (right_vertex, left_vertex)

# Returns the shape of the edge 'edge' of a tetrahedron having shape 'shape'.
# The edge is identified by the (unoriented) pair (first_vertex, second_vertex)
# We suppose that shape_error < min_imaginary (< shape.imag())

def get_edge_shape (edge, shape):
	one = 1
	two = ic.complex(itl.interval(2))
	if edge == (0,2) or edge == (2,0) or edge == (1,3) or edge == (3,1):
		shape = one / (one - shape)

	if edge == (0,3) or edge == (3,0) or edge == (1,2) or edge == (2,1):
		shape = one - one / shape

	return shape

# Returns the cusp shape z = x + yi of the cusp number num_cusp

def get_cusp_shape(N, num_cusp, tetrahedra_shapes):
	triangulation = get_triangulation(N)
	peripheral_curves = get_peripheral_curves(N)
	num_tetrahedra = len(triangulation)
	starting_edge = get_starting_edge(triangulation, peripheral_curves, num_cusp)
	translation = [0, 0]

	# Construct a euclidean triangulation on each cusp
	
	edge_length = get_euclidean_triangulation(N, triangulation, tetrahedra_shapes)

	for which_curve in range(2):
		
		# We always suppose that N is orientable and tetrahedra are glued with
		# orientation-reversing maps. That is, all the permutations 1023 are odd.
		#
		# Every tetrahedron is oriented. Every triangle of the cusp tori is also
		# correspondingly oriented. Its three edges are cyclically oriented. 
		# Triangles are glued orientation-reversely.

		out_tet = starting_edge[0]
		out_vertex = starting_edge[1]
		out_edge = starting_edge[2]

		# If the curve at the starting edge is not oriented coherently,
		# we invert its orientation and record this fact in 'invert_curve'

		invert_curve = False
		if peripheral_curves[which_curve][out_tet][out_vertex][out_edge] < 0:
			for t in range(num_tetrahedra):
				for v in range(4):
					for e in range(4):
						peripheral_curves[which_curve][t][v][e] = - peripheral_curves[which_curve][t][v][e] 
			invert_curve = True
		
		# We store the left and right vertex of the starting edge.
		# (The output pair is (right, left) instead of (left, right) because
		# we are exiting (out) from this triangle through this edge!)
		
		out_right_vertex, out_left_vertex = get_oriented_vertices(out_edge, out_vertex)
 
		# We begin by putting the starting edge on the complex plane 
		# as an oriented segment from 0 to 1. The curve may intersect this
		# edge multiple times, so we record in 'position' the actual intersection
		# starting from zero along the oriented edge.

		out_position = 1		

		# We now develop the triangulation along the curve on the complex plane. 
		# We will get a path of some N edges from 0 to some point z = a + ib 
		# which will represent the complex length the curve.
		
		left_point = 0
		right_point = edge_length[out_tet][out_vertex][out_edge]
		
		keep_going_on = True
		while keep_going_on:

			# The curve enters a new triangle

			in_tet = triangulation[out_tet][out_edge][0]
			in_vertex = triangulation[out_tet][out_edge][1][out_vertex]
			in_edge = triangulation[out_tet][out_edge][1][out_edge]
			in_left_vertex = triangulation[out_tet][out_edge][1][out_left_vertex]
			in_right_vertex = triangulation[out_tet][out_edge][1][out_right_vertex]
			in_position = out_position

			left_edge = in_right_vertex
			right_edge = in_left_vertex
			
			# The curve bends to the left or to the right?

			bend_left = True
			label_left_edge = peripheral_curves[which_curve][in_tet][in_vertex][left_edge]
			if label_left_edge < in_position:
				bend_left = False

			# The curve crosses the triangle			

			out_tet = in_tet
			out_vertex = in_vertex

			if bend_left:
				out_edge = left_edge
				out_left_vertex = in_left_vertex
				out_right_vertex = in_edge
				out_position = in_position
				right_point = right_point - edge_length[out_tet][out_vertex][out_left_vertex]

			else:
				out_edge = right_edge
				out_left_vertex = in_edge
				out_right_vertex = in_right_vertex
				out_position = in_position - label_left_edge

				left_point = left_point + edge_length[out_tet][out_vertex][out_right_vertex]

			new_edge = (out_tet, out_vertex, out_edge)
			if new_edge == starting_edge and out_position == 1:
				keep_going_on = False
		translation[which_curve] = left_point
		if invert_curve: translation[which_curve] = - translation[which_curve]
	cusp_shape = translation[1] / translation[0]
	return (cusp_shape.real, - cusp_shape.imag)
	
	
	